use if ! ( grep { $_ eq '.' } @INC ), qw[lib .];
use lib 'inc';
use Devel::AssertOS qw[Linux];

# This file was automatically generated by Dist::Zilla::Plugin::ModuleBuild v6.012.
use strict;
use warnings;

use Module::Build 0.28;


my %module_build_args = (
  "build_requires" => {
    "Module::Build" => "0.28"
  },
  "configure_requires" => {
    "ExtUtils::MakeMaker" => 0,
    "Module::Build" => "0.28"
  },
  "dist_abstract" => "wrapper around containers/buildah tool for multi-stage builds of OCI/Docker-compatible Linux containers",
  "dist_author" => [
    "Ian Kluft <https://github.com/ikluft>"
  ],
  "dist_name" => "Container-Buildah",
  "dist_version" => "0.2.0",
  "license" => "apache",
  "module_name" => "Container::Buildah",
  "recursive_test_files" => 1,
  "requires" => {
    "Algorithm::Dependency" => 0,
    "Algorithm::Dependency::Source::HoA" => 0,
    "Carp" => 0,
    "Class::Singleton" => 0,
    "Cwd" => 0,
    "Data::Dumper" => 0,
    "Exporter" => 0,
    "File::Slurp" => 0,
    "File::Sync" => 0,
    "File::stat" => 0,
    "FindBin" => 0,
    "Getopt::Long" => 0,
    "IO::Handle" => 0,
    "IPC::Run" => 0,
    "Modern::Perl" => 0,
    "Readonly" => 0,
    "Template" => 0,
    "YAML::XS" => 0,
    "autodie" => 0,
    "parent" => 0,
    "strict" => 0
  },
  "test_requires" => {
    "File::Basename" => 0,
    "File::Path" => 0,
    "Test::More" => 0,
    "Test::RequiresInternet" => 0,
    "warnings" => 0
  }
);


my %fallback_build_requires = (
  "File::Basename" => 0,
  "File::Path" => 0,
  "Module::Build" => "0.28",
  "Test::More" => 0,
  "Test::RequiresInternet" => 0,
  "warnings" => 0
);


unless ( eval { Module::Build->VERSION(0.4004) } ) {
  delete $module_build_args{test_requires};
  $module_build_args{build_requires} = \%fallback_build_requires;
}

my $build = Module::Build->new(%module_build_args);


$build->create_build_script;
